#include "IrrCompileConfig.h"

#ifdef _IRR_COMPILE_WITH_DIRECT3D_9_

#include "os.h"
#include "CD3D9Context.h"
#include "CD3D9Driver.h"
#include "CD3D9Utils.h"

//#include <d3dx9core.h>
//#include <dxerr.h>

namespace irr
{
namespace video
{


CD3D9Context::CD3D9Context(CD3D9Driver* driver, IDirect3DDevice9* d3dDevice, const SVideoContextConfig& config, const bool& external ) :
    IVideoContext(config, external)
    , pID3DDevice(d3dDevice)
    , VideoDriver( driver )
//    , D3D9(0)
    , SceneSourceRect(0)
    , WindowHandle(0)
    {
        _IRR_DEBUG_BREAK_IF( VideoDriver == 0);
    }



CD3D9Context::~CD3D9Context()
{

}

IDirect3DDevice9* CD3D9Context::getDevice() const
{
    return pID3DDevice;
}


bool CD3D9Context::setTarget(void * data)
{
    HWND windowHandle = reinterpret_cast<HWND>( data );
    if( windowHandle )
    {
        WindowHandle = windowHandle;
        return true;
    }

    return false;
}



CD3D9Driver * CD3D9Context::getVideoDriver() const
{
    return VideoDriver;
}



/** See GLContext */
bool CD3D9Context::beginScene( DWORD flags, SColor color, core::rect<s32>* sceneSourceRect )
{
    SceneSourceRect = sceneSourceRect;

//	if (!pID3DDevice)
//    {
//        os::Printer::log("Invalid pID3DDevice.", ELL_WARNING);
//        return false;
//    }
//


//	DWORD flags = 0;

//	if (clearBackBuffer)
//		flags |= D3DCLEAR_TARGET;
//
//	if ( clearZBuffer)
//		flags |= D3DCLEAR_ZBUFFER;

    // Move it
	if ( getConfig().Stencilbuffer)
    {
        // TODO reestablish
//		flags |= D3DCLEAR_STENCIL;
    }


	HRESULT hr;
	if (flags)
	{
		hr = pID3DDevice->Clear(
                          0,    // nb of rectangles in following array
                          NULL, // rectangles of areas to clear
                          flags, // method's behavior. You can specify a flag to clear a render-target surface, an associated depth buffer, the stencil buffer, or any combination of the three
                          color.color,
                          1.0,  // depth buffer clearing value
                          0     // stencil buffer clearing value
                          );

        core::stringc temp("DIRECT3D9 clear failed.");
//        temp.append( DXGetErrorDescription(hr) );

//	Try to clear either the depth buffer or the stencil buffer of a render target that does not have an attached depth buffer.
//Try to clear the stencil buffer when the depth buffer does not contain stencil data.

		if (FAILED(hr))
//		if ( hr != D3D_OK)
        {
//            temp.append( DXGetErrorDescription(hr) );

//			os::Printer::log( temp.c_str(), ELL_WARNING);
//            DXTRACE_MSG("Error happened");
//            DXTRACE_ERR_MSGBOX("Error happened",hr);

//            DXTRACE_ERR( "DIRECT3D9 clear failed.", hr );
            _IRR_DXTRACE_ERR( "DIRECT3D9 clear failed.", hr );
			return false;
        }
	}

	hr = pID3DDevice->BeginScene();
	if ( FAILED(hr) )
	{
            _IRR_DXTRACE_ERR( "DIRECT3D9 begin scene failed. It is likely you already called beginScene without its associated endScene.", hr );

//	    core::stringc temp;
//	    switch(hr)
//	    {
//        case  D3DERR_INVALIDCALL :
//            temp = "DIRECT3D9 begin scene failed. It is likely you already called beginScene without its associated endScene";
////            temp = "test";
//            // TODO
////            temp.append( DXGetErrorDescription(hr) );
//            break;
//        default:
//            break;
//	    };
//		os::Printer::log(temp.c_str(), ELL_WARNING);
		return false;
	}

	return true;

}



/** See GLContext */
HRESULT CD3D9Context::endScene( DWORD flags)
{


	HRESULT hr = pID3DDevice->EndScene();
//	D3DERR_INVALIDCALL
	if (FAILED(hr))
	{
		os::Printer::log("DIRECT3D9 end scene failed.", ELL_WARNING);
		return false;
	}

	RECT* srcRct = 0;
	RECT sourceRectData;

	// TODO
	if ( SceneSourceRect )
	{
		srcRct = &sourceRectData;
		sourceRectData.left = SceneSourceRect->UpperLeftCorner.X;
		sourceRectData.top = SceneSourceRect->UpperLeftCorner.Y;
		sourceRectData.right = SceneSourceRect->LowerRightCorner.X;
		sourceRectData.bottom = SceneSourceRect->LowerRightCorner.Y;
	}



	IDirect3DSwapChain9* swChain = 0;
	hr = pID3DDevice->GetSwapChain(0, &swChain);
    if( FAILED(hr))
    {
        //!
        _IRR_DXTRACE_ERR("GetSwapChain failed",hr);
    }
//        D3D_OK
	// TODO
//	DWORD flags = ( getConfig().HandleSRGB && (getVideoDriver()->Caps.Caps3 & D3DCAPS3_LINEAR_TO_SRGB_PRESENTATION)) ? D3DPRESENT_LINEAR_CONTENT : 0;

    // TODO move it
    #ifdef _DEBUG
    if(!WindowHandle)
    {
        os::Printer::log("Invalid window handle", ELL_WARNING);
    }
    #endif

	hr = swChain->Present(
//                       NULL
                       srcRct           //!< src rect (may be null)
                       , NULL           //!< dest rect
//                       , NULL
                       , WindowHandle   //!< either set to null or override default target
                       , NULL           //!< dirty region
                       , flags          //!<
                       );
    if(FAILED(hr))
    {
        _IRR_DXTRACE_ERR("Present failed",hr);
    }

	swChain->Release();
    return hr;
}


bool CD3D9Context::isCompatibleWithRenderWindow(IRenderWindow* ) const
{
    //!
    return true;
}


void CD3D9Context::releaseContext()
{
	if (pID3DDevice)
		pID3DDevice->Release();


//	if (pID3D)
//		pID3D->Release();
}

#if 0

IDirect3DPixelShader9* CD3D9Context::createPixelShader(const c8* pxsh)
{
	if (!pxsh)
    {
        os::Printer::log("Empty Pixel shader code", ELL_ERROR);
        return 0;
    }


    IDirect3DPixelShader9* PixelShader = 0;

	// compile shader

	LPD3DXBUFFER code = 0;
	LPD3DXBUFFER errors = 0;

	#ifdef _IRR_D3D_NO_SHADER_DEBUGGING

		// compile shader without debug info
		stubD3DXAssembleShader(pxsh, (UINT)strlen(pxsh), 0, 0, 0, &code, &errors);
	#else

		// compile shader and emitt some debug informations to
		// make it possible to debug the shader in visual studio

		static int irr_dbg_file_nr = 0;
		++irr_dbg_file_nr;
		char tmp[32];
		sprintf(tmp, "irr_d3d9_dbg_shader_%d.psh", irr_dbg_file_nr);

		FILE* f = fopen(tmp, "wb");
		fwrite(pxsh, strlen(pxsh), 1, f);
		fflush(f);
		fclose(f);

		stubD3DXAssembleShaderFromFile(tmp, 0, 0, D3DXSHADER_DEBUG, &code, &errors);

	#endif


	if (errors)
	{
		// print out compilation errors.
		os::Printer::log("Pixel shader compilation failed:", ELL_ERROR);
		os::Printer::log((c8*)errors->GetBufferPointer(), ELL_ERROR);

		if (code)
			code->Release();

		errors->Release();
		return 0;
	}

	if (FAILED(pID3DDevice->CreatePixelShader((DWORD*)code->GetBufferPointer(), &PixelShader)))
	{
		os::Printer::log("Could not create pixel shader.", ELL_ERROR);
		code->Release();
		return 0;
	}

	code->Release();
	return PixelShader;
}







IDirect3DVertexShader9*
CD3D9ShaderMaterialRenderer::createVertexShader(const char* vtxsh)
{

    IDirect3DVertexShader9* VertexShader = 0;

	if (!vtxsh)
    {

        os::Printer::log("Empty vertex code",ELL_ERROR);
		return 0;
	}

	// compile shader

	LPD3DXBUFFER code = 0;
	LPD3DXBUFFER errors = 0;

	#ifdef _IRR_D3D_NO_SHADER_DEBUGGING

		// compile shader without debug info
		stubD3DXAssembleShader(vtxsh, (UINT)strlen(vtxsh), 0, 0, 0, &code, &errors);

	#else

		// compile shader and emitt some debug informations to
		// make it possible to debug the shader in visual studio

		static int irr_dbg_file_nr = 0;
		++irr_dbg_file_nr;
		char tmp[32];
		sprintf(tmp, "irr_d3d9_dbg_shader_%d.vsh", irr_dbg_file_nr);

		FILE* f = fopen(tmp, "wb");
		fwrite(vtxsh, strlen(vtxsh), 1, f);
		fflush(f);
		fclose(f);

        stubD3DXAssembleShaderFromFile(tmp, 0, 0, D3DXSHADER_DEBUG, &code, &errors);

	#endif

	if (errors)
	{
		// print out compilation errors.
		os::Printer::log("Vertex shader compilation failed:", ELL_ERROR);
		os::Printer::log((c8*)errors->GetBufferPointer(), ELL_ERROR);

		if (code)
			code->Release();

		errors->Release();
		return 0;
	}

	if (!code || FAILED(pID3DDevice->CreateVertexShader((DWORD*)code->GetBufferPointer(), &VertexShader)))
	{
		os::Printer::log("Could not create vertex shader.", ELL_ERROR);
		if (code)
			code->Release();
		return 0;
	}

	code->Release();
	return VertexShader;
}

#endif



} // video
} // irr

#endif // _IRR_COMPILE_WITH_DIRECT3D_8_
